use convert_case::{Case, Casing};
use quote::{format_ident, quote};

pub fn derive(tokens: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let input: syn::DeriveInput = syn::parse_macro_input!(tokens);

    let items = if let syn::Data::Struct(syn::DataStruct {
        struct_token: _,
        fields: syn::Fields::Named(fields),
        semi_token: _,
    }) = input.data
    {
        fields
            .named
            .into_iter()
            .map(|f| (f.ident.unwrap(), f.ty))
            .collect::<Vec<_>>()
    } else {
        panic!("Can only derive Database on data structs with named fields!");
    };

    println!("items: {:?}", &items);

    let db_ident = input.ident;

    let items = items.iter().map(|field| {
        let item_combined_name =
            format_ident!("{}{}", db_ident, field.0.to_string().to_case(Case::Snake));
        let item_base_name = &field.0.to_string();
        let item_type = &field.1;

        quote! {
            struct #item_combined_name;
            impl crate::db::DatabaseItem for #item_combined_name {
                fn item_key() -> &'static str { #item_base_name }
                fn dependency_keys() -> &'static [&'static str] { &[] } // TODO

                fn accept_entity_visitor(visitor: &mut impl crate::entity::EntityVisitor) {
                    <#item_type as crate::db::DatabaseSpec>::accept_entity_visitor(visitor);
                }
            }
            v.visit::<#item_combined_name>();
        }
    });

    let out = quote! {
        impl crate::db::Database for #db_ident {
            fn accept_item_visitor(v: &mut impl crate::db::DatabaseItemVisitor) {
                #(#items)*
            }
        }
    }
    .into();
    println!("out: {}", out);
    out
}
