use convert_case::{Case, Casing};
use quote::{format_ident, quote};

pub fn derive(tokens: proc_macro::TokenStream) -> proc_macro::TokenStream {
    let input: syn::DeriveInput = syn::parse_macro_input!(tokens);

    let parts = if let syn::Data::Struct(syn::DataStruct {
        struct_token: _,
        fields: syn::Fields::Named(fields),
        semi_token: _,
    }) = input.data
    {
        fields
            .named
            .into_iter()
            .map(|f| (f.ident.unwrap(), f.ty))
            .collect::<Vec<_>>()
    } else {
        panic!("Can only derive Entity on data structs with named fields!");
    };

    let entity_ident = input.ident;

    let part_bodies = parts.iter().map(|part| {
        let part_combined_name = format_ident!(
            "{}{}",
            entity_ident,
            part.0.to_string().to_case(Case::Snake)
        );
        let part_base_name = &part.0.to_string();
        let part_type = &part.1;
        quote! {
            struct #part_combined_name;
            impl crate::entity::EntityPart for #part_combined_name {
                type Datum = #part_type;
                fn part_name() -> &'static str {
                    #part_base_name
                }
            }
            v.visit::<#part_combined_name>();
        }
    });

    let entity_name = entity_ident.to_string().to_case(Case::Snake);

    quote! {
        impl crate::entity::Entity for #entity_ident {
            fn entity_name() -> &'static str { #entity_name }
            fn accept_part_visitor(v: &mut impl crate::entity::EntityPartVisitor) {
                #(
                    #part_bodies
                );*
            }
        }
    }
    .into()
}
