/// Represents a data field in an Entity
pub trait EntityDatum: 'static {
    fn sql_type() -> &'static str;
}

impl EntityDatum for String {
    fn sql_type() -> &'static str {
        "text"
    }
}

/// A single data field in an Entity
pub trait EntityPart: 'static {
    type Datum: EntityDatum;

    fn part_name() -> &'static str;
    fn is_unique() -> bool {
        false
    }
    fn is_assoc() -> bool {
        false
    }
    fn assoc_name() -> &'static str {
        unreachable!()
    }
    fn visit_assoc(_: &mut impl EntityVisitor) {
        unreachable!()
    }
}

/// Visitor for traversing all `EntityPart`s in an `Entity`
pub trait EntityPartVisitor {
    fn visit<EP: EntityPart>(&mut self);
}

/// A single database entity, aka an object type that gets its own table
pub trait Entity: 'static {
    fn entity_name() -> &'static str;
    fn accept_part_visitor(visitor: &mut impl EntityPartVisitor);
}

/// Visitor for traversing all `Entity`s in a container
pub trait EntityVisitor {
    fn visit<E: Entity>(&mut self);
}
